"""
Module to handle configuration, such as data directory, etc.
"""
import configparser
from pathlib import Path
from erddapy import ERDDAP


_default_config = {
    "data_directory": str(Path(__file__).parent.joinpath("data/").absolute()),
    "erddap_server": "http://erddap.bio-oracle.org/erddap/",
    "skip_confirmation": False,
}

config_file = Path(__file__).absolute().parent.joinpath(("config.ini"))


def create_config(default_config: dict = _default_config, path: str = None) -> None:
    """
    Creates the configuration file.
    """
    config = configparser.ConfigParser()
    config["DEFAULT"] = default_config
    if config_file.exists():
        response = input(
            f"Config file '{config_file}' already exists, overwrite it? y/N \n"
        )
        if (response.lower() not in "y yes") or (not response):
            print("Operation cancelled.\n")
            return
    with open(config_file, "w") as f:
        config.write(f)
        print(f"Created configuration at '{config_file}'.")


def _get_default_config() -> configparser.ConfigParser:
    if config_file.exists():
        current_config = configparser.ConfigParser()
        _ = current_config.read(config_file)
        return current_config
    else:
        try:
            print("Config file doesn't exist, creating it.")
            create_config()
            return _get_default_config()
        except Exception as e:
            print(
                f"Error: could not load or create configuration file. Loading default values."
            )
            config = configparser.ConfigParser()
            config["DEFAULT"] = _default_config
            return config


def get_config_path() -> Path:
    """
    Returns the path to the config file.
    """
    if config_file.exists():
        return config_file
    else:
        print("Config file doesn't exist, creating it.")
        create_config()
        get_config_path()


def print_config_values() -> None:
    """
    Prints configuration values.
    """
    if not config_file.exists():
        print("Config file doesn't exist, creating it.")
        create_config()
    print(f"Configuration values in '{config_file}' are as following:\n")
    for key, value in _get_default_config()["DEFAULT"].items():
        print(key, "\t", value)
    print()
    print(
        f"Edit the configuration file '{config_file}' or use the 'update_setting' function in this module to edit them."
    )


def update_setting(key, value) -> None:
    """
    Modifies a setting in the configuration file.
    """
    config = _get_default_config()
    config["DEFAULT"][key] = value
    with open(config_file, "w") as f:
        config.write(f)
    print(f"Successfully updated config file at '{config_file}'.")


config = dict(_get_default_config()["DEFAULT"])
default_server = ERDDAP(
    server=config["erddap_server"],
)



"""
Utitilies and private methods that are used internally.
"""

from copy import deepcopy
from datetime import datetime
from functools import lru_cache
from glob import glob
import logging
from pathlib import Path

import httpx

from pyo_oracle.config import default_server, config


def convert_bytes(num):
    """
    this function will convert bytes to MB.... GB... etc

    from https://stackoverflow.com/questions/2104080/how-do-i-check-file-size-in-python
    """
    for x in ["bytes", "KB", "MB", "GB", "TB"]:
        if num < 1024.0:
            return "%3.1f %s" % (num, x)
        num /= 1024.0


def _format_args(function):
    """
    Converts list/str to tuple.

    This is used as we want functions to work on multiple items, but also want to allow users
    to pass strings.

    Prevent TypeError when passing list to function with lru_cache.

    From: https://stackoverflow.com/questions/49210801/python3-pass-lists-to-function-with-functools-lru-cache
    """

    def wrapper(*args, **kwargs):
        kwargs = {k: tuple(x) if type(x) == list else x for k, x in kwargs.items()}
        kwargs = {k: (x,) if type(x) == str else x for k, x in kwargs.items()}
        args = [tuple(x) if type(x) == list else x for x in args]
        args = [(x,) if type(x) == str else x for x in args]
        result = function(*args, **kwargs)
        return result

    return wrapper


def _validate_argument(name: str, value: str or list or tuple, valid_values: list):
    """Check if argument is in a valid list of values."""
    if value:
        for v in value:
            if v.lower() not in valid_values:
                print(
                    f"Selected {name} '{v}' is not a valid {name}. These are valid values:\n{valid_values}"
                )


def _download_file_from_url(url: str, local_path: Path):
    """
    Downloads a large file from a URL.
    From: https://stackoverflow.com/questions/16694907/download-large-file-in-python-with-requests
    """
    with open(local_path, "wb") as f:
        with httpx.stream("GET", url, follow_redirects=True) as response:
            for chunk in response.iter_bytes():
                f.write(chunk)
    return local_path


# Small helpers
verbose_print = lambda message, verbose: print(message) if verbose else None
info_logger = lambda msg, log: logging.info(msg) if log else None


def confirm(msg, cancel_msg="Download cancelled."):
    print(
        "You can disable these confirmation prompts by passing 'skip_confirmation=True' to the function,"
        " or set the 'skip_confirmation' setting to True in the config.ini file.\n"
    )
    response = input(msg + " Would you like to proceed? y/N\n")
    if response.lower() in "y yes".split():
        return True
    else:
        print(cancel_msg)
        return False


def _get_griddap_dataset_url(
    dataset_id,
    variables=None,
    constraints=None,
    response="nc",
    verbose=False,
):
    server = deepcopy(default_server)
    server.dataset_id = dataset_id
    server.protocol = "griddap"
    server.griddap_initialize()

    verbose_print(f"Selected '{dataset_id}' dataset.", verbose)
    verbose_print(f"Dataset info available at: {server.get_info_url()}", verbose)

    # Setting constraints that match
    if constraints:
        constraints = {**server._constraints_original, **constraints}
        for k, v in constraints.items():
            if k in server.constraints.keys():
                server.constraints[k] = v
                server._constraints_original[k] = v

    # The same for variables
    if variables:
        server.variables = [v for v in server.variables if v in variables]
    verbose_print(
        f"Selected {len(server.variables)} variables: {server.variables}.", verbose
    )
    url = server.get_download_url(response=response)
    return url


@lru_cache(2)
def _layer_dataframe(include_allDatasets=False):
    """
    Requests the list of layers as a pandas DataFrame.

        include_allDatasets (bool): whether to include the row with the 'allDatasets' dataset.
    """
    s = deepcopy(default_server)
    s.dataset_id = "allDatasets"
    s.protocol = "tabledap"
    df = s.to_pandas()

    if not include_allDatasets:
        df = df.iloc[1:]
    return df


def _download_layer(
    dataset_id: str or list,
    output_directory: str or Path = None,
    response: str = "nc",
    constraints: dict = None,
    skip_confirmation=None,
    verbose=True,
    log=True,
    timestamp=True,
):
    timestamp = datetime.now().strftime("%Y-%m-%d-%H%M%S")
    filename = (
        f"{dataset_id}_{timestamp}.{response}"
        if timestamp
        else f"{dataset_id}.{response}"
    )
    outdir = (
        Path(output_directory) if output_directory else Path(config["data_directory"])
    )
    local_path = outdir.joinpath(filename)

    # Check for existing layers
    if skip_confirmation is None:
        skip_confirmation = eval(config["skip_confirmation"])
    if not skip_confirmation:
        print()
        print(f"Data directory is '{outdir}'.", end="\n\n")
        existing_layers = [
            i
            for i in outdir.iterdir()
            if str(i.name).startswith(dataset_id) and not str(i.name).endswith(".log")
        ]
        if existing_layers:
            print(
                f"You have {len(existing_layers)} local datasets with the same prefix as '{dataset_id}' in the data directory:"
            )
            for layer in existing_layers:
                print(layer.name, "\t", convert_bytes(layer.stat().st_size))
            print()
            print("You can check it again by running `list_local_data()`.")
            print(
                "To check the constraints of each existing dataset, see the log files in the data directory.",
                end="\n\n",
            )

            msg = ""
            if not confirm(msg):
                return

    # Configure logging
    logfile = local_path.with_suffix(".log")
    logging.basicConfig(
        filename=logfile,
        format="%(asctime)s %(message)s",
        level=logging.INFO,
        datefmt="%Y-%m-%d %H%M%S",
    )

    msg = f"Downloading dataset '{dataset_id}' as {response} file to '{outdir}'."
    info_logger(msg, log)
    verbose_print(msg, verbose)

    msg = "Constraints are:"
    info_logger(msg, log)
    verbose_print(msg, verbose)
    info_logger(constraints, log)
    verbose_print(constraints, verbose)

    url = _get_griddap_dataset_url(
        dataset_id, constraints=constraints, response=response
    )  # Response is converted to
    _download_file_from_url(url, local_path)
    if (verbose or log) and local_path.exists():
        filesize = convert_bytes(local_path.stat().st_size)
        msg = f"Download finished at '{local_path}'. File size is {filesize}."
        info_logger(msg, log)
        verbose_print(msg, verbose)



from pathlib import Path

from erddapy import ERDDAP
import pandas as pd
import pytest

import pyo_oracle as pyo


@pytest.fixture(scope="session")
def test_data_dir(tmp_path_factory):
    test_data_dir = tmp_path_factory.mktemp("test_dir")
    return test_data_dir


@pytest.fixture
def constraints():
    constraints = {
    "time>=": "2000-01-01T00:00:00Z",
    "time<=": "2010-01-01T12:00:00Z",
    "time_step": 100,
    "latitude>=": 0,
    "latitude<=": 10,
    "latitude_step": 100,
    "longitude>=": 0,
    "longitude<=": 10,
    "longitude_step": 1
    }
    return constraints


@pytest.fixture
def layer():
    layer = "thetao_baseline_2000_2019_depthsurf"
    return layer


def test_list_layers():
    # Simple call
    layers_df_all = pyo.list_layers()
    assert isinstance(layers_df_all, pd.DataFrame)
    assert layers_df_all.empty is False

    # Test variables
    layers_df_filter = pyo.list_layers(variables=["po4",])
    assert isinstance(layers_df_filter, pd.DataFrame)
    assert layers_df_filter.empty is False
    assert len(layers_df_filter) < len(layers_df_all)

    # Test ssp
    layers_df_filter = pyo.list_layers(ssp=["ssp119",])
    assert isinstance(layers_df_filter, pd.DataFrame)
    assert layers_df_filter.empty is False
    assert len(layers_df_filter) < len(layers_df_all)

    # Test time
    layers_df_filter = pyo.list_layers(time_period="present")
    assert isinstance(layers_df_filter, pd.DataFrame)
    assert layers_df_filter.empty is False
    assert len(layers_df_filter) < len(layers_df_all)

    # Test list
    layers_list = pyo.list_layers(variables="po4", ssp=["ssp119", "ssp126"], time_period="future", dataframe=False)
    assert isinstance(layers_list, list)
    assert len(layers_list) > 0


def test_download_layers(layer, constraints, test_data_dir):
    pyo.download_layers(layer, output_directory=test_data_dir, response="csv", constraints=constraints, skip_confirmation=True)


def test_list_local_data():
    pyo.list_local_data(test_data_dir, )
    pyo.list_local_data(test_data_dir, verbose=True)


def test_manual_erddapy_requests(layer, constraints, test_data_dir):
    e = ERDDAP(server=pyo.config["erddap_server"], protocol="griddap")
    e.dataset_id = layer
    e.griddap_initialize()
    e.constraints = constraints
    e._constraints_original = constraints
    df = e.to_pandas()
    assert isinstance(df, pd.DataFrame)
    assert df.empty is False
    outfile = test_data_dir.joinpath(f"{layer}.csv")
    df.to_csv(outfile)
    assert Path(outfile).exists()



"""
Main module with library functions.
"""
from functools import lru_cache
from glob import glob
from pathlib import Path

import pandas as pd

from pyo_oracle.config import config
from pyo_oracle.utils import (
    _format_args,
    _download_layer,
    _validate_argument,
    _layer_dataframe,
    verbose_print,
    convert_bytes,
    confirm,
)


def download_layers(
    dataset_ids: str or list,
    output_directory: str or Path = None,
    response: str = "nc",
    constraints: dict = None,
    skip_confirmation=None,
    verbose=True,
    log=True,
    timestamp=True,
) -> None:
    """
    Downloads one or more layers.
    """
    if isinstance(dataset_ids, str):
        dataset_ids = (dataset_ids,)

    if skip_confirmation is None:
        skip_confirmation = eval(config["skip_confirmation"])

    if not skip_confirmation and not constraints:
        question = "No constraints have been set. This will download the full dataset, which may be a few GBs in size."

        if not confirm(question):
            return

    for dataset_id in dataset_ids:
        _download_layer(
            dataset_id,
            output_directory,
            response,
            constraints,
            skip_confirmation,
            verbose,
            log,
            timestamp,
        )


@_format_args
@lru_cache(8)
def list_layers(
    variables: str or list = None,
    ssp: str or list = None,
    time_period: str = None,
    dataframe: bool = True,
    _include_allDatasets: bool = False,
) -> pd.DataFrame or list:
    """
    Lists available layers in the Bio-ORACLE server.

        variables (str|list): variables to filter from. Valid values are ['po4','o2','si','ph','sws','phyc','so','thetao','dfe','no3','sithick','tas','siconc','chl','mlotst','clt','terrain'].
        ssp (str|list): future scenario to choose from. Valid values are ['ssp119', 'ssp126', 'ssp370', 'ssp585', 'ssp460', 'ssp245', 'baseline'].
        time_period (str): time period to choose from. Valie values are either 'present' or 'future'.
        dataframe (bool): whether to return a Pandas DataFrame. If False, will return a list.
    """
    valid_variables = [
        "po4",
        "o2",
        "si",
        "ph",
        "sws",
        "phyc",
        "so",
        "thetao",
        "dfe",
        "no3",
        "sithick",
        "tas",
        "siconc",
        "chl",
        "mlotst",
        "clt",
        "terrain",
    ]
    valid_ssp = ["ssp119", "ssp126", "ssp370", "ssp585", "ssp460", "ssp245", "baseline"]
    valid_time_period = "present future".split()

    for arg in "variables", "ssp", "time_period":
        _validate_argument(arg, eval(arg), eval(f"valid_{arg}"))

    _dataframe = _layer_dataframe(_include_allDatasets)
    # Filter resulting dataframe
    if variables:
        _dataframe = pd.concat(
            [
                _dataframe[_dataframe["datasetID"].str.startswith(v.lower())]
                for v in variables
            ]
        )
    if time_period:
        start_decade = _dataframe["datasetID"].str.split("_", expand=True)[2]
        if time_period == ("present",):
            _dataframe = _dataframe[start_decade == "2000"]
        elif time_period == ("future",):
            _dataframe = _dataframe[start_decade == "2020"]
    if ssp:
        _dataframe = pd.concat(
            [
                _dataframe[
                    (_dataframe["datasetID"].str.contains(v.lower()))
                    | (_dataframe["datasetID"].str.contains(v.capitalize()))
                    | (_dataframe["datasetID"].str.contains(v.upper()))
                ]
                for v in ssp
            ]
        )

    # Convert to list
    if dataframe:
        return _dataframe.reset_index(drop=True)
    else:
        return _dataframe["datasetID"].to_list()


def list_local_data(data_directory=None, verbose=True):
    """
    Lists datasets that are locally downloaded.
    """
    if data_directory is None:
        data_directory = config["data_directory"]

    verbose_print(f"Your data directory is '{data_directory}'.\n", verbose)
    verbose_print("Contents of data directory:", verbose)
    files = glob(str(Path(config["data_directory"]).joinpath("*")))
    if not verbose:
        files = [f for f in files if not str(f).endswith(".log")]
    if files:
        for file in files:
            print(
                "\t", Path(file).name, "\t", convert_bytes(Path(file).stat().st_size)
            ) if verbose else print("\t", Path(file).name)
        print()
    else:
        print(f"Data directory at '{data_directory}' does not contain any data.")
    if verbose:
        dirsize = sum(Path(f).stat().st_size for f in files)
        dirsize = convert_bytes(dirsize)
        verbose_print(f"Size of data directory is {dirsize}.", verbose)
